using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Collections.Specialized;
using System.Web;
using System.IO;
using System.Xml;
using System.Text.RegularExpressions;
using System.Globalization;

namespace Gcalendar
{
    public class GoogleAuth
    {
        private string email, password;
        private string servicecode, authtoken, errorcode;

        const string authUrl = "https://www.google.com/accounts/ClientLogin";
        const string appName = "PHL-GcalClient-1.0";

        public string AuthToken
        {
            get { return authtoken; }
            private set { authtoken = value; }
        }

        public string ErrorCode
        {
            get { return errorcode; }
            private set { errorcode = value; }
        }

        public GoogleAuth(string i, string p, string s)
        {
            email = i;
            password = p;
            servicecode = s;
        }

        public string Email
        {
            get { return email; }
        }

        public string Password
        {
            get { return password; }
        }

        public string ServiceCode
        {
            get { return servicecode; }
        }

        private WebRequest PostRequest(string url, NameValueCollection sendData, Encoding textEncoding)
        {
            WebRequest myRequest = WebRequest.Create(url);
            myRequest.Method = "POST";
            myRequest.ContentType = "application/x-www-form-urlencoded";

            string requestString = "";

            for (int i = 0; i < sendData.Count; i++)
            {
                if (i != 0)
                    requestString += "&";
                requestString +=
                    HttpUtility.UrlEncode(sendData.Keys[i], textEncoding) + "=" +
                    HttpUtility.UrlEncode(sendData[i], textEncoding);
            }

            StreamWriter myStreamWriter = new
                StreamWriter(myRequest.GetRequestStream(), Encoding.ASCII);
            myStreamWriter.Write(requestString);
            myStreamWriter.Close();

            return myRequest;
        }

        public bool RequestAuth()
        {

            NameValueCollection authParameter = new NameValueCollection();

            authParameter.Add("Email", email);
            authParameter.Add("Passwd", password);
            authParameter.Add("service", servicecode);
            authParameter.Add("source", appName);

            WebRequest myRequest = PostRequest(authUrl, authParameter, Encoding.ASCII);
            HttpWebResponse myResponse = null;

            try
            {
                myResponse = (HttpWebResponse)myRequest.GetResponse();
            }
            catch (WebException ex)
            {
                if (ex.Status == WebExceptionStatus.ProtocolError)
                    myResponse = (HttpWebResponse)ex.Response;
            }

            if (myResponse == null)
                return false;

            StreamReader sr = new StreamReader(myResponse.GetResponseStream());

            string buffer = null;

            while ((buffer = sr.ReadLine()) != null)
            {
                if (buffer.StartsWith("Auth="))
                    AuthToken = buffer.Substring(5);
                if (buffer.StartsWith("Error="))
                    ErrorCode = buffer.Substring(6);
            }

            myResponse.Close();

            return (myResponse.StatusCode == HttpStatusCode.OK);
        }
    }

    public class EventFeed : GoogleFeed
    {
        private List<EventEntry> feedList;
        
        public EventFeed(string myUrl, GoogleAuth myAuth) : base (myUrl, myAuth)
        {
        }

        public List<EventEntry> Query()
        {
            feedList = new List<EventEntry>();
            GetFeed();
            return feedList;
        }

        protected override void ParseNode(XmlNode entryNode, XmlNamespaceManager xmlNs)
        {
            EventEntry myEntry = new EventEntry();

            myEntry.Title = GetNodeValue(entryNode, "atom:title/text()", xmlNs);
            myEntry.Visibility = GetNodeValue(entryNode, "gd:visibility/@value", xmlNs);
            myEntry.Where = GetNodeValue(entryNode, "gd:where/@valueString", xmlNs);

            string stTime = GetNodeValue(entryNode, "gd:when/@startTime", xmlNs);
            string endTime = GetNodeValue(entryNode, "gd:when/@endTime", xmlNs);

            if (stTime != null)
            {
                myEntry.IsAllDay = (stTime.IndexOf("T") == -1);
                myEntry.StartTime = ParseDateTime(stTime);
                myEntry.EndTime = myEntry.StartTime;
            }

            if (endTime != null)
                myEntry.EndTime = ParseDateTime(endTime);

            feedList.Add(myEntry);
        }
    }

    public class CalendarFeed : GoogleFeed
    {
        private List<CalendarEntry> feedList;
        
        public CalendarFeed(string myUrl, GoogleAuth myAuth)
            : base(myUrl, myAuth)
        {
        }

        public List<CalendarEntry> Query()
        {
            feedList = new List<CalendarEntry>();
            GetFeed();
            return feedList;
        }

        protected override void ParseNode(XmlNode entryNode, XmlNamespaceManager xmlNs)
        {
            CalendarEntry myEntry = new CalendarEntry();

            myEntry.Title = GetNodeValue(entryNode, "atom:title/text()", xmlNs);
            myEntry.FeedUrl = GetNodeValue(entryNode, @"atom:link/@href[../@rel=""alternate""]", xmlNs);

            feedList.Add(myEntry);
        }
    }
    
    public abstract class GoogleFeed
    {
        private string feedUrl;
        private XmlDocument returnXml;
        private GoogleAuth feedAuth;

        private static string authToken, sessionId;
        private static GoogleAuth prevAuth;

        protected abstract void ParseNode(XmlNode entryNode, XmlNamespaceManager xmlNs);
        
        public string FeedUrl
        {
            get { return feedUrl; }
            set { feedUrl = value; }
        }

        public string FeedText
        {
            get
            {
                if (FeedDoc != null)
                    return FeedDoc.OuterXml;
                else
                    return null;
            }
        }

        protected GoogleFeed(string myUrl, GoogleAuth myAuth)
        {
            FeedUrl = myUrl;
            feedAuth = myAuth;
        }

        private XmlDocument FeedDoc
        {
            get { return returnXml; }
        }

        protected void GetFeed()
        {

            if (prevAuth == null || feedAuth == null || prevAuth.Email != feedAuth.Email || prevAuth.Password != feedAuth.Password || prevAuth.ServiceCode != feedAuth.ServiceCode)
            {
                authToken = null;
                sessionId = null;
            }

            int tryCount = 2;

            do
            {
                try
                {
                    DownloadFeed();
                    tryCount = 0;
                }
                catch (Exception ex)
                {
                    authToken = null;
                    if (--tryCount == 0)
                        throw ex;
                }
            } while (tryCount > 0);

            XmlNamespaceManager xmlNs = new XmlNamespaceManager(FeedDoc.NameTable);

            xmlNs.AddNamespace("atom", "http://www.w3.org/2005/Atom");
            xmlNs.AddNamespace("gd", "http://schemas.google.com/g/2005");
            xmlNs.AddNamespace("gCal", "http://schemas.google.com/gCal/2005");

            XmlNodeList entryList = FeedDoc.SelectNodes("//atom:entry", xmlNs);

            foreach (XmlNode currentNode in entryList)
            {
                ParseNode(currentNode, xmlNs);
            }
        }

        private void DownloadFeed()
        {
            if (authToken == null && feedAuth != null)
            {
                if (feedAuth.RequestAuth())
                {
                    authToken = feedAuth.AuthToken;
                    prevAuth = feedAuth;
                }
                else
                    throw (new ApplicationException("F؂Ɏs܂"));
            }

            string getUrl = feedUrl;

            if (sessionId != null)
            {
                getUrl += (feedUrl.IndexOf("?") > -1 ? "&" : "?") + "gsessionid=" + HttpUtility.UrlEncode(sessionId);
            }

            bool isFailure = false;

            HttpWebResponse myResponse = null;
            int code;

            do
            {
                HttpWebRequest myRequest = (HttpWebRequest)(WebRequest.Create(getUrl));
                myRequest.AllowAutoRedirect = false;
                if (feedAuth != null)
                    myRequest.Headers.Add("Authorization: GoogleLogin auth=" + authToken);

                myRequest.UserAgent = "GcalClient/1.00";

                try
                {
                    myResponse = (HttpWebResponse)myRequest.GetResponse();
                }
                catch (WebException ex)
                {
                    if (ex.Status == WebExceptionStatus.ProtocolError)
                        myResponse = (HttpWebResponse)ex.Response;
                }
                code = (int)myResponse.StatusCode;

                if (code >= 300 && code < 400)
                {
                    sessionId = null;
                    getUrl = myResponse.Headers["Location"];
                }

                if (code >= 400 && code < 500)
                {
                    if (sessionId != null)
                    {
                        sessionId = null;
                        getUrl = feedUrl;
                    }
                    else
                    {
                        isFailure = true;
                    }
                }

                returnXml = null;
                if (myResponse != null)
                {
                    Stream sr = myResponse.GetResponseStream();
                    if (sr != null)
                    {
                        returnXml = new XmlDocument();
                        returnXml.Load(sr);
                    }

                    myResponse.Close();
                }

            } while ((isFailure == false) && (code >= 300 && code < 500));

            if (isFailure)
            {
                throw (new ApplicationException("f[^̎擾Ɏs܂" + code));
            }
            else
            {
                sessionId = ParseSessionId(getUrl);
            }
        }

        private string ParseSessionId(string url)
        {
            Regex session = new Regex("[?&]gsessionid=([^&]*)");
            Match myMatch = session.Match(url);
            if (myMatch.Success)
                return myMatch.Groups[1].Value;
            else
                return null;
        }

        protected DateTime ParseDateTime(string s)
        {
            string[] dateFormat;
            DateTimeStyles dtStyles = DateTimeStyles.AllowWhiteSpaces;

            if (s.EndsWith("Z"))
            {
                dateFormat = new string[1] { "yyyy-MM-ddTHH:mm:ss.FFFZ" };
                dtStyles |= DateTimeStyles.AssumeUniversal;
            }
            else
            {
                dateFormat = new string[2] { "yyyy-MM-ddTHH:mm:ss.FFFzzz", "yyyy-MM-dd"};
                dtStyles |= DateTimeStyles.AssumeLocal;
            }
            DateTime d = DateTime.ParseExact(s, dateFormat, CultureInfo.InvariantCulture, dtStyles);

            return d;
        }

        protected string GetNodeValue(XmlNode node, string xpath, XmlNamespaceManager ns)
        {
            XmlNode searchNode = node.SelectSingleNode(xpath, ns);
            if (searchNode == null)
                return null;
            else
                return searchNode.Value;
        }
    }

    public class CalendarEntry
    {
        private string title, feed;

        public string Title
        {
            get { return title; }
            set { title = value; }
        }

        public string FeedUrl
        {
            get { return feed; }
            set { feed = value; }
        }
    }

    public class EventVisibility
    {
        public const string Confidential = "http://schemas.google.com/g/2005#event.confidential";
        public const string Default = "http://schemas.google.com/g/2005#event.default";
        public const string Private = "http://schemas.google.com/g/2005#event.private";
        public const string Public = "http://schemas.google.com/g/2005#event.public";
    }

    public class EventEntry
    {
        private string title, where, visibility;
        private bool isAllDay;
        private DateTime? starttime, endtime;

        public string Title
        {
            get { return title; }
            set { title = value; }
        }

        public DateTime? StartTime
        {
            get { return starttime; }
            set { starttime = value; }
        }

        public DateTime? EndTime
        {
            get { return endtime; }
            set { endtime = value; }
        }

        public bool IsAllDay
        {
            get { return isAllDay; }
            set { isAllDay = value; }
        }

        public string Visibility
        {
            get { return visibility; }
            set { visibility = value; }
        }

        public string Where
        {
            get { return where; }
            set { where = value; }
        }
    }
}
